import logging
import socket
import time

from kafka3 import KafkaProducer
from kafka3.errors import KafkaTimeoutError
from .common import CURRENT_PROD_BROKER_VERSION

log = logging.getLogger("woof")


def make_kafka_safe(raw_data):
    """
    This function was written to avoid non-unicode
    string data produced to Kafka
    """
    return raw_data


class TransactionLogger(object):
    def __init__(self,
                 broker,
                 vertical,
                 host=socket.gethostname(),
                 async_args=False,
                 retries=1,
                 key_serializer=make_kafka_safe,
                 value_serializer=make_kafka_safe,
                 **kwargs):
        self.broker = broker
        self.this_host = host
        self.vertical = vertical
        self.async_args = async_args
        self.topic = _get_topic_from_vertical(vertical)
        kwargs['api_version'] = kwargs.get('api_version',
                                           CURRENT_PROD_BROKER_VERSION)
        # thread safe producer, uses default murmur2 partiioner by default
        # good for us
        self.producer = KafkaProducer(bootstrap_servers=broker,
                                      key_serializer=key_serializer,
                                      value_serializer=value_serializer,
                                      retries=retries,
                                      **kwargs)

    def New(self,
            txn_id,
            amount,
            skus,
            detail="#",
            userid="#",
            email="#",
            phone="#"):
        self._send_log("NEW", txn_id, amount, skus, detail, userid, email,
                       phone)

    def Modify(self,
               txn_id,
               amount="#",
               skus=[],
               detail="#",
               userid="#",
               email="#",
               phone="#"):
        self._send_log("MODIFY", txn_id, amount, skus, detail, userid, email,
                       phone)

    def Cancel(self,
               txn_id,
               amount="#",
               skus=[],
               detail="#",
               userid="#",
               email="#",
               phone="#"):
        self._send_log("CANCEL", txn_id, amount, skus, detail, userid, email,
                       phone)

    def Fulfil(self,
               txn_id,
               amount="#",
               skus=[],
               detail="#",
               userid="#",
               email="#",
               phone="#"):
        self._send_log("FULFIL", txn_id, amount, skus, detail, userid, email,
                       phone)

    def _send_log(self,
                  verb,
                  txn_id,
                  amount,
                  skus,
                  detail="#",
                  userid="#",
                  email="#",
                  phone="#",
                  retry=True,
                  retry_time_in_s=1):
        msg = self._format_message(verb, txn_id, amount, skus, detail, userid,
                                   email, phone)
        log.info("[transactions log] topic %s txnid %s msg %s /n", self.topic,
                 txn_id, msg)
        try:
            future = self.producer.send(self.topic, key=txn_id, value=msg)
            future.get(timeout=1)
        except KafkaTimeoutError as e:
            log.error(
                "[transactions log] KafkaTimeoutError ERROR %s topic %s txnid %s msg %s /n",
                str(e), self.topic, txn_id, msg)
            if retry:
                time.sleep(retry_time_in_s)
                self._send_log(verb,
                               txn_id,
                               amount,
                               skus,
                               detail,
                               userid,
                               email,
                               phone,
                               retry=False)
            else:
                raise e
        except Exception as e1:
            log.error(
                "[transactions log] GEN error ERROR %s topic %s txnid %s msg %s /n",
                str(e1), self.topic, txn_id, msg)
            raise e1

    def _format_message(self, verb, txn_id, amount, skus, detail, userid,
                        email, phone):
        """
        Generates log message.
        """
        separator = '\t'

        safe_skus = [make_kafka_safe(x) for x in skus]
        skus_as_string = ",".join(safe_skus)

        return separator.join([self.this_host, str(time.time(
        )), verb, make_kafka_safe(txn_id), make_kafka_safe(
            amount), skus_as_string, make_kafka_safe(detail), make_kafka_safe(
                userid), make_kafka_safe(email), make_kafka_safe(phone)])


def _get_topic_from_vertical(vertical):
    return "_".join(["TRANSACTIONS", vertical])



import threading, logging, time, signal

from kafka3 import KafkaConsumer
from kafka3.errors import KafkaTimeoutError
from .common import WoofNotSupported, CURRENT_PROD_BROKER_VERSION

log = logging.getLogger("woof")


class FeedConsumer(threading.Thread):
    """Threaded gomsg feed consumer
    callbacks are called on a separate thread in the same method
    
    keyword arguments :

    broker (list): List of initial broker nodes the consumer should contact to
    bootstrap initial cluster metadata.  This does not have to be the full node list.
    It just needs to have at least one broker

    group (str): the name of the consumer group to join, Offsets are fetched /
    committed to this group name.

    offset='earliest' : read all msgs from beginning of time;  default read fresh

    async=True : In case of Zk commit offset everytime

    commit_every_t_ms:  How much time (in milliseconds) to before commit to zookeeper

    kwargs : anything else you want to pass to kafka-python

    """
    daemon = True

    def __init__(self,
                 broker,
                 group,
                 offset='latest',
                 commit_every_t_ms=1000,
                 parts=None,
                 kill_signal=signal.SIGTERM,
                 wait_time_before_exit=1,
                 use_zk=False,
                 async_commit=True,
                 handler_timeout_ms=30000,
                 **kwargs):
        self.brokerurl = broker
        self.kill_signal = kill_signal
        self.exit_consumer = False
        self.async_commit = async_commit
        try:
            self.create_kill_signal_handler()
        except Exception as e:
            log.error(
                "[feedconsumer log] exception %s. Skipping signal handler install. ",
                str(e))
            pass

        self.wait_time_before_exit = wait_time_before_exit

        if use_zk:
            kwargs['api_version'] = kwargs.get('api_version', '0.8.1')
            # ZK autocommit does not seem to work reliably
            # TODO
            self.async_commit = False

        else:
            kwargs['api_version'] = kwargs.get('api_version',
                                               CURRENT_PROD_BROKER_VERSION)

        # curb over-optimism
        handler_timeout_ms = min(handler_timeout_ms, 60000)

        try:
            self.cons = KafkaConsumer(
                bootstrap_servers=broker,
                auto_offset_reset=offset,
                enable_auto_commit=self.async_commit,
                auto_commit_interval_ms=commit_every_t_ms,
                group_id=group,
                session_timeout_ms=handler_timeout_ms,
                **kwargs)

        except KafkaTimeoutError as e:
            log.error(
                "[feedconsumer log] INIT KafkaTimeoutError  %s. Please check broker string %s \n",
                str(e), broker)
            raise e
        except Exception as e1:
            log.error("[feedconsumer log] INIT err %s \n", str(e1))
            raise e1

        self.callbacks = {}
        super(FeedConsumer, self).__init__()

    def add_topic(self, topic, todo, parts=None):
        """
        Set the topic/partitions to consume

        todo (callable) : callback for the topic
        NOTE: Callback is for entire topic, if you call this for multiple
        partitions for same topic with diff callbacks, only the last callback
        is retained

        topic : topic to listen to

        parts (list) : tuple of the partitions to listen to

        """
        try:
            self.callbacks[topic] = todo

            if parts is None:
                log.info("[feedconsumer log] : adding topic %s ", topic)
            else:
                raise WoofNotSupported(
                    "manual partition assignement not supported")

            self.cons.subscribe(topics=self.callbacks.keys())
        except Exception as e:
            log.error("[feedconsumer log] add_topic err %s \n", str(e))
            raise e

    def remove_topic(self, topic, parts=None):
        if parts is not None:
            raise WoofNotSupported(
                "manual partition assignement not supported")

        try:
            self.cons.unsubscribe()
            del self.callbacks[topic]
            self.cons.subscribe(topics=self.callbacks.keys())
        except Exception as e:
            log.error("[feedconsumer log] remove_topic err %s \n", str(e))
            raise e

    def create_kill_signal_handler(self):
        def set_stop_signal(signal, frame):
            self.exit_consumer = True

        signal.signal(self.kill_signal, set_stop_signal)

    def check_for_exit_criteria(self):
        if self.exit_consumer:
            self.cons.commit()
            time.sleep(self.wait_time_before_exit)
            exit(0)

    def run(self):
        while True:
            try:
                for m in self.cons:
                    self.callbacks[m.topic](m.key, m.value)
                    # Looks like in kakfa-python,  the generator namely (fetcher._message_generator), stores offset in
                    # the next call to __next__
                    # https://github.com/dpkp/kafka-python/blob/master/kafka/consumer/fetcher.py
                    # TODO verfiy

                    # "extra" safely
                    if not self.async_commit:
                        self.cons.commit()

                    self.check_for_exit_criteria()
                self.check_for_exit_criteria()
            except Exception as e:
                log.error(
                    "[feedconsumer log] thread run  err %s ..continuing..\n",
                    str(e))
                time.sleep(1)



import logging
import random

from kafka3 import KafkaProducer
from kafka3.errors import KafkaTimeoutError
from kafka3.partitioner.default import DefaultPartitioner
from .common import CURRENT_PROD_BROKER_VERSION
from .transactions import make_kafka_safe

log = logging.getLogger("woof")

BATCH_SEND_DEFAULT_INTERVAL = 20
BATCH_SEND_MSG_COUNT = 32
REQUEST_TIMEOUT_MS = 2000
MAX_BLOCK_MS = 1000


class PartitionedProducer(object):
    """
    use send() to send to any topic and distribute based on key
    """

    def __init__(self, broker,
                 partitioner=None,  # Note if the earlier hash is needed, need to explicitly pass dumb_hash
                 async_args=False,
                 req_acks=None,  # unused  - here for legacy support
                 ack_timeout=None,  # unused  - here for legacy support
                 codec=None,
                 batch_send=False,
                 batch_send_every_n=BATCH_SEND_MSG_COUNT,
                 batch_send_every_t=BATCH_SEND_DEFAULT_INTERVAL,  # unused  - here for legacy support
                 retries=3,
                 key_serializer=make_kafka_safe,
                 value_serializer=make_kafka_safe,
                 **kwargs):

        try:
            self.async_args = async_args
            if partitioner is not None:
                _partitioner = CustomPartitioner(partitioner)
            else:
                _partitioner = DefaultPartitioner()
            kwargs['api_version'] = kwargs.get('api_version',
                                               CURRENT_PROD_BROKER_VERSION)
            self.prod = KafkaProducer(bootstrap_servers=broker,
                                      key_serializer=key_serializer,
                                      value_serializer=value_serializer,
                                      batch_size=batch_send_every_n,
                                      retries=retries,
                                      partitioner=_partitioner,
                                      request_timeout_ms=REQUEST_TIMEOUT_MS,
                                      max_block_ms=MAX_BLOCK_MS,
                                      **kwargs)
        except Exception as e1:
            log.error("[partitionedproducer log] GEN err %s  \n", str(e1))
            raise

    def send(self, topic, key, *msg):
        try:
            for _msg in msg:
                self.prod.send(topic, key=key, value=_msg)

            # for async flush will happen in background
            if not self.async_args:
                self.prod.flush()

        except KafkaTimeoutError as e:
            log.error(
                "[feedproducer log] KafkaTimeoutError err %s topic %s  \n",
                str(e), topic)
            raise e
        except Exception as e1:
            log.error("[feedproducer log] GEN  err %s topic %s \n", str(e1),
                      topic)
            raise e1


# Note if the earlier hash is needed, need to explicitly pass dumb_hash
def dumb_hash(key):
    sum = 0
    str_key = str(key)
    for s in str_key:
        sum += ord(s)

    log.debug("[feedproducer log] dumb_hash , key = %s", sum)
    return sum


class CustomPartitioner(object):
    _hash_map = {}

    def __init__(self, hasher):
        CustomPartitioner._hash_map[1] = hasher

    @classmethod
    def __call__(cls, key, all_partitions, available):

        if key is None:
            if available:
                return random.choice(available)
            return random.choice(all_partitions)

        idx = cls._hash_map[1](key)
        idx &= 0x7fffffff
        idx %= len(all_partitions)
        return all_partitions[idx]


class CyclicPartitionedProducer(KafkaProducer):
    """
    use send() to send to any topic and distribute keys cyclically in partitions
    """

    def __init__(self,
                 broker,
                 async_args=True,
                 key_serializer=make_kafka_safe,
                 value_serializer=make_kafka_safe,
                 random_start=True,
                 **kwargs):
        self.partition_cycles = {}
        self.random_start = random_start
        self.async_args = async_args
        kwargs['api_version'] = kwargs.get('api_version',
                                           CURRENT_PROD_BROKER_VERSION)
        super(CyclicPartitionedProducer, self).__init__(
            bootstrap_servers=broker,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            **kwargs)

    def _partition(self, topic, partition, key, value, serialized_key,
                   serialized_value):
        if partition is not None:
            assert partition >= 0
            assert partition in self._metadata.partitions_for_topic(
                topic), 'Unrecognized partition'
            return partition

        all_partitions = list(self._metadata.partitions_for_topic(topic))
        n_partitions = len(all_partitions)

        try:
            offset = (self.partition_cycles[topic] + 1) % n_partitions
        except:
            if self.random_start:
                offset = random.randint(0, n_partitions - 1)
            else:
                offset = 0

        self.partition_cycles[topic] = offset
        return all_partitions[offset]

    def send(self, topic, key, *msg):
        try:
            for _msg in msg:
                super(CyclicPartitionedProducer, self).send(topic,
                                                            key=key,
                                                            value=_msg)

            # for async flush will happen in background
            if not self.async_args:
                self.prod.flush()

        except KafkaTimeoutError as e:
            log.error(
                "[feedproducer log] KafkaTimeoutError err %s topic %s  \n",
                str(e), topic)
            raise e
        except Exception as e1:
            log.error("[feedproducer log] GEN  err %s topic %s \n", str(e1),
                      topic)
            raise e1



from gevent import monkey

monkey.patch_all()
import logging, gevent, time
import signal
import sys, traceback

from .consumer import FeedConsumer
from .common import WoofNotSupported

log = logging.getLogger("woof")


class GreenFeedConsumer(FeedConsumer):
    """
    Greenlet based feed consumer
    All callbacks are spawned as greenlets on a background thread


    keyword arguments :

    broker (list): List of initial broker nodes the consumer should contact to
    bootstrap initial cluster metadata.  This does not have to be the full node list.
    It just needs to have at least one broker

    group (str): the name of the consumer group to join, Offsets are fetched /
    committed to this group name.

    offset='smallest' : read all msgs from beginning of time;  default read fresh

    commit_every_t_ms:  How much time (in milliseconds) to before commit to zookeeper
    kill_signal: What is the kill signal to handle exit gracefully.
    wait_time_before_exit: How much time to wait before exiting green threads

    """
    daemon = True

    def __init__(self,
                 broker,
                 group,
                 offset='largest',
                 commit_every_t_ms=1000,
                 parts=None,
                 kill_signal=signal.SIGTERM,
                 wait_time_before_exit=1):

        # TODO
        # getting issues with gevent and kafka-python (select to be specific)
        # https://github.com/dpkp/kafka-python/issues/702
        # unsupported till them
        raise WoofNotSupported("GreenFeedConsumer  not supported")

        super(GreenFeedConsumer, self).__init__(
            broker, group, offset, commit_every_t_ms, parts, kill_signal,
            wait_time_before_exit)

    def wrap(self, callback, mesg):
        # earlier wrap was used to mark task done, but not needed here i guess
        # Looks like in kakfa-python,  the generator namely (fetcher._message_generator), stores offset in
        # the next call to __next__
        # https://github.com/dpkp/kafka-python/blob/master/kafka/consumer/fetcher.py
        # TODO this may make greenlet based callbacks dangerous..as we have commited without callback returning
        callback(mesg.key, mesg.value)

    def run(self):
        while True:
            try:
                for m in self.cons:
                    gevent.spawn(self.wrap, self.callbacks[m.topic], m)
                    self.check_for_exit_criteria()
                self.check_for_exit_criteria()
            except Exception as e:
                print('EXCEPTION', e)
                traceback.print_exc(file=sys.stdout)
                log.error(
                    "[greenconsumer log] thread run  err %s ..continuing../n",
                    str(e))
                time.sleep(1)



import logging, time

from kafka3 import KafkaProducer
from kafka3.errors import KafkaTimeoutError
from .common import CURRENT_PROD_BROKER_VERSION
from .transactions import make_kafka_safe

log = logging.getLogger("woof")


class FeedProducer(object):
    """
    Feed Producer class
    use send() to send to any topic
    """

    def __init__(self,
                 broker,
                 key_serializer=make_kafka_safe,
                 value_serializer=make_kafka_safe,
                 retries=3,
                 async_arg=False,
                 **kwargs):
        try:
            kwargs['api_version'] = kwargs.get('api_version',
                                               CURRENT_PROD_BROKER_VERSION)
            self.prod = KafkaProducer(bootstrap_servers=broker,
                                      key_serializer=key_serializer,
                                      value_serializer=value_serializer,
                                      retries=retries,
                                      **kwargs)
            self.async_arg = async_arg
        except Exception as e:
            log.error("[feedproducer log] Constructor error ERROR %s  \n",
                      str(e))
            raise

    def send(self, topic, *msgs):
        try:
            for msg in msgs:
                future = self.prod.send(topic, msg)
                log.info(
                    "[feedproducer log] about to flush.. recordmeta %s message %s \n",
                    str(future.get(timeout=1)),
                    msg)

        except KafkaTimeoutError as e:
            log.error(
                "[feedproducer log] KafkaTimeoutError err %s topic %s message %s \n",
                str(e), topic, str(msgs))
            raise e
        except Exception as e1:
            log.error("[feedproducer log] GEN  err %s topic %s message %s \n",
                      str(e1), topic, str(msgs))
            raise e1



import sys
import os
from setuptools import setup, Command


here = os.path.abspath(os.path.dirname(__file__))

with open('VERSION', 'r') as v:
    __version__ = v.read().rstrip()

with open(os.path.join(here, 'README.md')) as f:
    README = f.read()





setup(
    name="woof3",
    version=__version__,

    packages=[
        "woof"
    ],

    author="Jyotiswarup Raiturkar",
    author_email="jyotisr5@gmail.com",
    description="Messaging library ",
    long_description='',
    keywords="apache kafka",
    install_requires=['six', 'kafka-python', 'gevent'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)
